function [ resultarray, timearray ] = ExactHeterodyneSolver( F,U,Delta,gamma,psiinit,timearray)
%EXACTHETERODYNESOLVER Solving full trajectories of true fock-base states, in a
%vectorized way.
%generally columns:components; rows: samples; Third dimension( only in resultarray):
%timestep.
%F,U,Delta,gamma: physical parameters of the dynamics as defined in the main text
%psiinit: matrix of initial states
%timearray: row vector of the times when results must be stored
%resultarray: wavefunction components for the different samples at
%different timesteps


%For a coherent initial state with mean-field value alpha, psiinit can be
%constructed trough
%a=sparse(Nmax,Nmax);for it=1:Nmax-1,  a(it,it+1)=sqrt(it); end;adag=a'; %annihilation and creation operators for use in the displacement operator.
%vacuumpsi=zeros(Nmax,Nmax); vacuumpsi(1,:)=1; % state of Vacuum
%psiinit=exp(-0.5*abs(alpha)^2)*(expm(alpha*adag)*vacuumpsi);% Coherent
%state (see e.g. Formula (2.2.3) in Scully-Zubairy)

%by W. Verstraelen, Theory of Quantum Systems & Complex Systems, dept.
%Physics, University of Antwerp for MATLAB 2017b.
%developed for 'Gaussian Quantum Trajectories for the Variational
%Simulation of Open Quatum-Optical Systems--WV and M. Wouters'


alternativeexpression1=false; %update rules that can numerically a bit more stable under some circumstances.
alternativeexpression2=false;
if(alternativeexpression1&&alternativeexpression2),error('incompatible booleans'); end
dt=1e-6;%1e-7;

%rng(10); %random seed can be set if wanted


Nmax=size(psiinit,1);
Nsamples=size(psiinit,2);



    
    a=sparse(Nmax,Nmax);for it=1:Nmax-1,  a(it,it+1)=sqrt(it); end %#ok<SPRIX>
    adag=a';
    n=adag*a; 

    unit=speye(Nmax);

    H=-Delta* n +F*adag+conj(F)*a+U/2*n.*(n-unit); 


    if(alternativeexpression1)
        fixedpart=expm(-1i*dt*(H-0.5i*gamma*n));
        detprefactor=1;
    elseif(alternativeexpression2)
        fixedpart=0;
        detprefactor=expm(-1i*dt*(H-0.5i*gamma*n));
    else
        fixedpart=unit-1i*dt*(H-0.5i*gamma*n);
        detprefactor=1;
    end
    scaleda=gamma*dt*a;

    resultarray=zeros(Nmax,Nsamples,length(timearray));

    sqdt=sqrt(dt);

    time=timearray(1);
    resultarray(:,:,1)=psiinit;
    psimatprev=psiinit;
    while time<timearray(end)

    dW1=sqdt*randn(1,Nsamples); dW1=repmat(dW1,[Nmax 1]); 
    dW2=sqdt*randn(1,Nsamples); dW2=repmat(dW2,[Nmax 1]); 

    adagexp=diag(psimatprev'*adag*psimatprev); % dimension [Nsamples Nsamples]
    adagexpmat=repmat(adagexp.',[Nmax 1]); % matrix of the same size as psimat, containing expectation values;

    psimat=fixedpart*psimatprev+scaleda*(adagexpmat.*psimatprev)+a*sqrt(0.5*gamma)*((dW1-1i*dW2).*psimatprev); %Met minteken is ruis consistent met de andere methodes bij zelfde random numbers
    psimat=detprefactor*psimat;
       norms = sqrt(sum(abs(psimat).^2,1)); % Compute norms of columns (https://nl.mathworks.com/matlabcentral/answers/280464-normalizing-columns-of-a-matrix)
       psimat = bsxfun(@rdivide,psimat,norms); 

     time=time+dt;
     step=find(abs(timearray-time)<dt/2);

            if ~isempty(step)
                resultarray(:,:,step)=psimat;
                assert(all(isfinite(psimat(:))));
                %[dW1(1) dW2(1)]/sqdt ->uncomment for the random numbers to
                %be printed
            end

    psimatprev=psimat;
    end


end

