function [ resultarray, timearray ] = ExactPhotonCountingSolver( F,U,Delta,gamma,psiinit,timearray)
%EXACTPHOTONCOUNTINGSOLVER Solving full trajectories of true fock-base states, in a
%vectorized way.
%generally columns:components; rows: samples; Third dimension( only in resultarray):
%timestep.
%F,U,Delta,gamma: physical parameters of the dynamics as defined in the main text
%psiinit: matrix of initial states
%timearray: row vector of the times when results must be stored
%resultarray: wavefunction components for the different samples at
%different timesteps


%For a coherent initial state with mean-field value alpha, psiinit can be
%constructed trough
%a=sparse(Nmax,Nmax);for it=1:Nmax-1,  a(it,it+1)=sqrt(it); end;adag=a'; %annihilation and creation operators for use in the displacement operator.
%vacuumpsi=zeros(Nmax,Nmax); vacuumpsi(1,:)=1; % state of Vacuum
%psiinit=exp(-0.5*abs(alpha)^2)*(expm(alpha*adag)*vacuumpsi);% Coherent
%state (see e.g. Formula (2.2.3) in Scully-Zubairy)

%by W. Verstraelen, Theory of Quantum Systems & Complex Systems, dept.
%Physics, University of Antwerp for MATLAB 2017b.
%developed for 'Gaussian Quantum Trajectories for the Variational
%Simulation of Open Quatum-Optical Systems--WV and M. Wouters'






dt=1e-06; %timestep



     Nmax=size(psiinit,1); %Max particle number considered, size of Hilbert space
    Nsamples=size(psiinit,2); %Amount of independent realizations
   
    

    
    a=sparse(Nmax,Nmax);for it=1:Nmax-1,  a(it,it+1)=sqrt(it); end %#ok<SPRIX>
    adag=a';
    n=adag*a; 

    unit=speye(Nmax);

    H=-Delta* n +F*adag+conj(F)*a+U/2*n.*(n-unit); 
    J=sqrt(gamma)*a;%Jump-operator
    propagator=expm(-1i*dt*(H-0.5i*gamma*n));

    
    resultarray=zeros(Nmax,Nsamples,length(timearray));


    time=timearray(1);
    resultarray(:,:,1)=psiinit;
    psimat=psiinit;
    
     R=rand(1,Nsamples);%Initialize the values of norm when there will be a jump    

    
    while time<timearray(end)

        %Is there a jump?

        normsqd = sum(abs(psimat).^2,1);
        jumppresent=(normsqd<R); nojumppresent=~jumppresent;
    
        jumppresentmat=repmat(jumppresent,[Nmax  1]); nojumppresentmat=~jumppresentmat;
        R=nojumppresent.*R+jumppresent.*rand(1,Nsamples);
        
     
        %execution of jump
        psimat_jump=J*psimat; 
        jumpednorms=sqrt(sum(abs(psimat_jump).^2,1));
        psimat_jump=bsxfun(@rdivide,psimat_jump,jumpednorms);
        
        psimat=nojumppresentmat.*psimat+jumppresentmat.*psimat_jump;
       
        %det. evolution.
        psimat=propagator*psimat;
        
     time=time+dt;
     step=find(abs(timearray-time)<dt/2);

            if ~isempty(step)
                norms = sqrt(sum(abs(psimat).^2,1));              
                resultarray(:,:,step)=bsxfun(@rdivide,psimat,norms);
                assert(all(isfinite(psimat(:))));
            end
    end


      
end