function [Nresultarray,DNDNresultarray,PHresultarray,DPHDPHresultarray,DNDPHSYMresultarray,timearray]=NP_GaussianHeterodyneSolver(F,U,Delta,gamma,Ninit,DNDNinit,PHinit,DNDPHSYMinit,timearray)

%NP_GaussianHeterodyneSolver evolves trajectories with an N\Theta-Gaussian
%ansatz in a vectorized way.
%F,U,Delta,gamma: physical parameters of the dynamics as defined in the main text
%Ninit: column vector of initial densities
%DNDNinit: column vector of initial density variances
%PHinit: column vector of initial phases
%DNDPHSYMinit: column vector of the initial symmetric covariances of density and phase
%timearray: row vector of the times when results must be stored
%Nresultarray: density values of different samples at different timesteps
%DNDNresultarray: density variances of different samples at different timesteps
%PHresultarray: phases of different samples at different timesteps
%DPHDPHresultarray: phase variances of different samples at different
%timesteps
%DNDPHSYMresultarray: symmetric density-phase covariances of different
%samples of different timesteps

%An initial coherent state can (if density is not too low) be
%constructed by imposing Poisson statistics (DNDNinit=Ninit) and
%DNDPHSYMinit=0;

%by W. Verstraelen, Theory of Quantum Systems & Complex Systems, dept.
%Physics, University of Antwerp for MATLAB 2017b.
%developed for 'Gaussian Quantum Trajectories for the Variational
%Simulation of Open Quatum-Optical Systems--WV and M. Wouters'


dt=1e-6;%1e-5;%1e-4;
%rng(10);
withGPU=false; %perform the calculation on the GPU?

Xdecayfraction=0.5; %Fraction of the decay for which X or P quadrature is measured: Heterodyne corresponds with 50/50 case.
Pdecayfraction=0.5; assert(Xdecayfraction+Pdecayfraction==1);

assert(iscolumn(Ninit));%The different samples are ordered in the column dimension here.
assert(all(size(Ninit)==size(DNDNinit)&size(Ninit)==size(PHinit)&size(Ninit)==size(DNDPHSYMinit)));
vecsize=length(Ninit);


Nresultarray=zeros(vecsize,length(timearray)); 
DNDNresultarray=zeros(vecsize,length(timearray)); 
PHresultarray=zeros(vecsize,length(timearray)); 
DPHDPHresultarray=zeros(vecsize,length(timearray));
DNDPHSYMresultarray=zeros(vecsize,length(timearray));

if(withGPU)

time=timearray(1);
    n=gpuArray(Ninit);
    dndn=gpuArray(DNDNinit);
    ph=gpuArray(PHinit);
    dndphsym=gpuArray(DNDPHSYMinit);
    dphdph=(dndphsym.^2+0.25)./dndn;
    Nresultarray(:,1)=gather(n);
    DNDNresultarray(:,1)=gather(dndn);
    PHresultarray(:,1)=gather(ph)
    DPHDPHresultarray(:,1)=gather(dphdph);
    DNDPHSYMresultarray(:,1)=gather(dndphsym);

    sqdt=sqrt(dt);
    unit=gpuArray.ones(vecsize,1);
    sqgamotwo=sqrt(0.5*gamma);
    while time<timearray(end)

 
        dWX=sqdt*gpuArray.randn(vecsize,1);
        dWP=sqdt*gpuArray.randn(vecsize,1);
        dZ=sqrt(Xdecayfraction)*dWX+1i*sqrt(Pdecayfraction)*dWP;
        
        d_n_det=2*imag(F*C1())-gamma*n;
        d_dndn_det=4*imag(F*C2()-2*imag(F*C1))-2*gamma.*dndn+gamma.*n-2*gamma.*abs(C2()-C1()).^2;
        d_ph_det=Delta+U*(0.5-n)-real(F*C3());
        d_dndphsym_det=-U*dndn+2*imag(F*C6())-real(F*C7())-gamma*dndphsym-2*gamma.*real((C2()-C1()).*conj(C6()));
        
        d_n_stoch=2*sqrt(gamma)*real((C2()-C1()).*dZ);
        d_dndn_stoch=2*sqrt(gamma)*real((D3()-2*C2()+C1().*(1-dndn)).*dZ);
        d_ph_stoch=2*sqrt(gamma)*real(C6().*dZ);
        d_dndphsym_stoch=2*sqrt(gamma)*real((-C6()-(dndphsym+0.5i).*C1()+D2()).*dZ);
        
        n=n+d_n_det.*dt+d_n_stoch;
        dndn=dndn+d_dndn_det.*dt+d_dndn_stoch;
        ph=ph+d_ph_det.*dt+d_ph_stoch;
        dndphsym=dndphsym+d_dndphsym_det.*dt+d_dndphsym_stoch;
         dphdph=(dndphsym.^2+0.25)./dndn;

        time=time+dt;

        step=find(abs(timearray-time)<dt/2);
        if ~isempty(step)
      Nresultarray(:,step)=gather(n);
            DNDNresultarray(:,step)=gather(dndn);
            PHresultarray(:,step)=gather(ph);
            DPHDPHresultarray(:,step)=gather(dphdph);
            DNDPHSYMresultarray(:,step)=gather(dndphsym);
            assert(all(isfinite(n(:))));
        end

    end

else
    
   time=timearray(1);
    n=Ninit;
    dndn=DNDNinit;
    ph=PHinit;
    dndphsym=DNDPHSYMinit;
    dphdph=(dndphsym.^2+0.25)./dndn;
    Nresultarray(:,1)=n;
    DNDNresultarray(:,1)=dndn;
    PHresultarray(:,1)=ph;
    DPHDPHresultarray(:,1)=dphdph;
    DNDPHSYMresultarray(:,1)=dndphsym;

    sqdt=sqrt(dt);
    while time<timearray(end)

 
        dWX=sqdt*randn(vecsize,1);
        dWP=sqdt*randn(vecsize,1);
        dZ=sqrt(Xdecayfraction)*dWX+1i*sqrt(Pdecayfraction)*dWP;
        
        d_n_det=2*imag(F*C1())-gamma*n;
        d_dndn_det=4*imag(F*C2()-2*imag(F*C1))-2*gamma.*dndn+gamma.*n-2*gamma.*abs(C2()-C1()).^2;
        d_ph_det=Delta+U*(0.5-n)-real(F*C3());
        d_dndphsym_det=-U*dndn+2*imag(F*C6())-real(F*C7())-gamma*dndphsym-2*gamma.*real((C2()-C1()).*conj(C6()));
        
        d_n_stoch=2*sqrt(gamma)*real((C2()-C1()).*dZ);
        d_dndn_stoch=2*sqrt(gamma)*real((D3()-2*C2()+C1().*(1-dndn)).*dZ);
        d_ph_stoch=2*sqrt(gamma)*real(C6().*dZ);
        d_dndphsym_stoch=2*sqrt(gamma)*real((-C6()-(dndphsym+0.5i).*C1()+D2()).*dZ);
        
        n=n+d_n_det.*dt+d_n_stoch;
        dndn=dndn+d_dndn_det.*dt+d_dndn_stoch;
        ph=ph+d_ph_det.*dt+d_ph_stoch;
        dndphsym=dndphsym+d_dndphsym_det.*dt+d_dndphsym_stoch;
         dphdph=(dndphsym.^2+0.25)./dndn;

        time=time+dt;
        %saving the results if necessary
        step=find(abs(timearray-time)<dt/2);
        if ~isempty(step)
      Nresultarray(:,step)=n;
            DNDNresultarray(:,step)=dndn;
            PHresultarray(:,step)=ph;
            DPHDPHresultarray(:,step)=dphdph;
            DNDPHSYMresultarray(:,step)=dndphsym;
            assert(all(isfinite(n(:))));
        end

    end

end

%% The correlators (see Appendix B in the main text)

function[c1]= C1()
c1=sqrt(n).*exp(-1i*ph-0.5*dphdph).*(1+1./(2*n).*(0.5-1i.*dndphsym)-dndn./(8.*n.^2));
    end

    function[c2]=C2()
c2=sqrt(n).*exp(-1i.*ph-0.5.*dphdph).*(0.5-1i.*dndphsym+1./(2.*n).*(dndn+0.25-0.5i.*dndphsym-dndphsym.^2)+3./(8*n.^2).*(-0.5.*dndn+1i.*dndphsym.*dndn));
    end

    function[c3]=C3()
c3=exp(-1i*ph-0.5*dphdph)./sqrt(n).*(1+1/(2.*n)*(-0.5+1i*dndphsym)+3*dndn./(8.*n.^2));        
    end

    function[c4]=C4()
 c4=exp(-1i*ph-0.5*dphdph)./sqrt(n).*(-1i*dphdph-1./(2.*n).*(dndphsym+0.5i)*(1-dphdph)-3i./(8.*n.^2).*(2.*(dndphsym+0.5i).^2+dndn.*dphdph));        
    end

    function[c5]=C5()
c5=n.^(-3/2)*exp(-1i*ph-0.5*dphdph).*(1+1/(2.*n).*(-1.5+3i*dndphsym)+15*dndn./(8.*n.^2));
    end

    function[c6]=C6()        
     c6=sqrt(n).*exp(-1i*ph-0.5*dphdph).*(-1i*dphdph+1./(2.*n).*(dndphsym+0.5i).*(1-dphdph)+1i./(8.*n.^2).*(2.*(dndphsym+0.5i).^2+dndn.*dphdph));        
    end

    function[c7]=C7()
c7=sqrt(n).*exp(-1i*ph-0.5*dphdph).*(1./(2.*n).*(1-2i*dndphsym)-dndn./(2.*n.^2));        
    end  

    function [d1]=D1()
 d1=sqrt(n).*exp(-1i*ph-0.5*dphdph).*((1-dphdph).*dphdph-1i./(2.*n).*(3.*dphdph-dphdph.^2).*(dndphsym+0.5i)-1./(8.*n.^2).*(dndn.*dphdph+2.*(dndphsym+0.5i).^2));
    end

    function [d2]=D2()
 d2=sqrt(n).*exp(-1i*ph-0.5*dphdph).*((dndphsym+0.5i).*(1-dphdph)-0.5i./n.*(dndn.*dphdph+(dndphsym+0.5i).^2.*(2-dphdph))-3./(8.*n.^2).*dndn.*(dndphsym+0.5i));      
    end

    function [d3]=D3()
 d3=sqrt(n).*exp(-1i*ph-0.5*dphdph).*(dndn-(dndphsym+0.5i).^2-0.5i./n.*(3.*dndn.*(dndphsym+0.5i)-6.*(dndphsym+0.5i).^3)-3.*dndn./(8.*n.^2));
    end

end