function [Nresultarray,DNDNresultarray,PHresultarray,DPHDPHresultarray,DNDPHSYMresultarray,timearray]=NP_GaussianPhotonCountingSolver(F,U,delta,gamma,Ninit,DNDNinit,PHinit,DNDPHSYMinit,timearray)
%NP_PhotonCountingSolver evolves trajectories with an N\Theta-Gaussian
%ansatz in a vectorized way.
%F,U,Delta,gamma: physical parameters of the dynamics as defined in the main text
%Ninit: column vector of initial densities
%DNDNinit: column vector of initial density variances
%PHinit: column vector of initial phases
%DNDPHSYMinit: column vector of the initial symmetric covariances of density and phase
%timearray: row vector of the times when results must be stored
%Nresultarray: density values of different samples at different timesteps
%DNDNresultarray: density variances of different samples at different timesteps
%PHresultarray: phases of different samples at different timesteps
%DPHDPHresultarray: phase variances of different samples at different
%timesteps
%DNDPHSYMresultarray: symmetric density-phase covariances of different
%samples of different timesteps

%An initial coherent state can (if density is not too low) be
%constructed by imposing Poisson statistics (DNDNinit=Ninit) and
%DNDPHSYMinit=0;

%by W. Verstraelen, Theory of Quantum Systems & Complex Systems, dept.
%Physics, University of Antwerp for MATLAB 2017b.
%developed for 'Gaussian Quantum Trajectories for the Variational
%Simulation of Open Quatum-Optical Systems--WV and M. Wouters'


dt=1e-6;%1e-5;%1e-4
%rng(1);
withGPU=false; %perform the calculation on the GPU?



assert(iscolumn(Ninit));%The different samples are ordered in the column dimension here.
assert(all(size(Ninit)==size(DNDNinit)&size(Ninit)==size(PHinit)&size(Ninit)==size(DNDPHSYMinit)));
vecsize=length(Ninit);



Nresultarray=zeros(vecsize,length(timearray)); 
DNDNresultarray=zeros(vecsize,length(timearray)); 
PHresultarray=zeros(vecsize,length(timearray)); 
DPHDPHresultarray=zeros(vecsize,length(timearray));
DNDPHSYMresultarray=zeros(vecsize,length(timearray));


if(withGPU) 

 time=timearray(1);
    n=gpuArray(Ninit);
    dndn=gpuArray(DNDNinit);
    ph=gpuArray(PHinit);
    dndphsym=gpuArray(DNDPHSYMinit);
    dphdph=(dndphsym.^2+0.25)./dndn;
    Nresultarray(:,1)=gather(n);
    DNDNresultarray(:,1)=gather(dndn);
    PHresultarray(:,1)=gather(ph);
    DPHDPHresultarray(:,1)=gather(dphdph);
    DNDPHSYMresultarray(:,1)=gather(dndphsym);

    
    unit=gpuArray.ones(vecsize,1);%ones(vecsize,1);
    norm=unit;
 
    R=gpuArray.rand(vecsize,1);%Initialize the values of norm when there will be a jump    

   
    while time<timearray(end)

       
        
        %Is there a jump?
        jumppresent=(norm<R); nojumppresent=~jumppresent;
        
       
        
        R=nojumppresent.*R+jumppresent.*gpuArray.rand(vecsize,1);
        %R=nojumppresent.*R+jumppresent.*Rnew;
        
        
        n_jump=n-1+dndn./n;
        dndn_jump=dndn.*(1-dndn./(n.^2));
        ph_jump=ph+dndphsym./n;
        dndphsym_jump=dndphsym.*(1-dndn./(n.^2));
        norm_jump=unit;
        
        n=nojumppresent.*n+jumppresent.*n_jump;
        dndn=nojumppresent.*dndn+jumppresent.*dndn_jump;
        ph=nojumppresent.*ph+jumppresent.*ph_jump;
        dndphsym=nojumppresent.*dndphsym+jumppresent.*dndphsym_jump;
        norm=nojumppresent.*norm+jumppresent.*norm_jump;
        dphdph=(dndphsym.^2+0.25)./dndn;

        %Regardless of whether a jump happened, this is the deterministic
        %evolution.        
        
        if(F==0)
          d_n=-gamma*dndn;
          d_dndn=0;
          d_ph=delta+U*(0.5-n)-gamma*dndphsym;
          d_dndphsym=-U*dndn;
          d_norm=-gamma.*n.*norm;
        else
          d_n=-gamma*dndn+2*imag(F*C1());
          d_dndn=0+4*imag(F*C2())-2*imag(F*C1());
          d_ph=delta+U*(0.5-n)-gamma*dndphsym-real(F*C3());
          d_dndphsym=-U*dndn+2*imag(F*C6())-real(F*C7());
          d_norm=-gamma.*n.*norm;             
        end
        
        
        n=n+d_n*dt;
        dndn=dndn+d_dndn*dt;
        ph=ph+d_ph*dt;
        dndphsym=dndphsym+d_dndphsym*dt;
        dphdph=(dndphsym.^2+0.25)./dndn;
        norm=norm+d_norm*dt;

        time=time+dt;

        step=find(abs(timearray-time)<dt/2);
        if ~isempty(step)
            Nresultarray(:,step)=gather(n);
            DNDNresultarray(:,step)=gather(dndn);
            PHresultarray(:,step)=gather(ph);
            DPHDPHresultarray(:,step)=gather(dphdph);
            DNDPHSYMresultarray(:,step)=gather(dndphsym);
            assert(all(isfinite(n(:))));
        end

    end
   
else
    
    time=timearray(1);
    n=Ninit;
    dndn=DNDNinit;
    ph=PHinit;
    dndphsym=DNDPHSYMinit;
    dphdph=(dndphsym.^2+0.25)./dndn;
    Nresultarray(:,1)=n;
    DNDNresultarray(:,1)=dndn;
    PHresultarray(:,1)=ph;
    DPHDPHresultarray(:,1)=dphdph;
    DNDPHSYMresultarray(:,1)=dndphsym;

    
    unit=ones(vecsize,1);%ones(vecsize,1);
    norm=unit;
 
    R=rand(vecsize,1);%Initialize the values of norm when there will be a jump    
    %Rnew=0;%Temporary
   
    while time<timearray(end)

       
        
        %Is there a jump?
        jumppresent=(norm<R); nojumppresent=~jumppresent;
        
       
        
        R=nojumppresent.*R+jumppresent.*rand(vecsize,1);
          
        
        
        
        n_jump=n-1+dndn./n;
        dndn_jump=dndn.*(1-dndn./(n.^2));
        ph_jump=ph+dndphsym./n;
        dndphsym_jump=dndphsym.*(1-dndn./(n.^2));
        norm_jump=unit;
        
        n=nojumppresent.*n+jumppresent.*n_jump;
        dndn=nojumppresent.*dndn+jumppresent.*dndn_jump;
        ph=nojumppresent.*ph+jumppresent.*ph_jump;
        dndphsym=nojumppresent.*dndphsym+jumppresent.*dndphsym_jump;
        norm=nojumppresent.*norm+jumppresent.*norm_jump;
        dphdph=(dndphsym.^2+0.25)./dndn;

        %Regardless of whether a jump happened, this is the deterministic
        %evolution.
        if(F==0)
          d_n=-gamma*dndn;
          d_dndn=0;
          d_ph=delta+U*(0.5-n)-gamma*dndphsym;
          d_dndphsym=-U*dndn;
          d_norm=-gamma.*n.*norm;
        else
          d_n=-gamma*dndn+2*imag(F*C1());
          d_dndn=0+4*imag(F*C2())-2*imag(F*C1());
          d_ph=delta+U*(0.5-n)-gamma*dndphsym-real(F*C3());
          d_dndphsym=-U*dndn+2*imag(F*C6())-real(F*C7());
          d_norm=-gamma.*n.*norm;             
        end
        n=n+d_n*dt;
        dndn=dndn+d_dndn*dt;
        ph=ph+d_ph*dt;
        dndphsym=dndphsym+d_dndphsym*dt;
        dphdph=(dndphsym.^2+0.25)./dndn;
        norm=norm+d_norm*dt;

        time=time+dt;

        step=find(abs(timearray-time)<dt/2);
        if ~isempty(step)
            Nresultarray(:,step)=n;
            DNDNresultarray(:,step)=dndn;
            PHresultarray(:,step)=ph;
            DPHDPHresultarray(:,step)=dphdph;
            DNDPHSYMresultarray(:,step)=dndphsym;
            assert(all(isfinite(n(:))));
        end

    end
    
    %jumptimes
end
    
  %% The correlators

   function[c1]= C1()
c1=sqrt(n).*exp(-1i*ph-0.5*dphdph).*(1+1./(2*n).*(0.5-1i.*dndphsym)-dndn./(8.*n.^2));
    end

    function[c2]=C2()
c2=sqrt(n).*exp(-1i.*ph-0.5.*dphdph).*(0.5-1i.*dndphsym+1./(2.*n).*(dndn+0.25-0.5i.*dndphsym-dndphsym.^2)+3./(8*n.^2).*(-0.5.*dndn+1i.*dndphsym.*dndn));
    end

    function[c3]=C3()
c3=exp(-1i*ph-0.5*dphdph)./sqrt(n).*(1+1/(2.*n)*(-0.5+1i*dndphsym)+3*dndn./(8.*n.^2));        
    end

    function[c4]=C4()
 c4=exp(-1i*ph-0.5*dphdph)./sqrt(n).*(-1i*dphdph-1./(2.*n).*(dndphsym+0.5i)*(1-dphdph)-3i./(8.*n.^2).*(2.*(dndphsym+0.5i).^2+dndn.*dphdph));        
    end

    function[c5]=C5()
c5=n.^(-3/2)*exp(-1i*ph-0.5*dphdph).*(1+1/(2.*n).*(-1.5+3i*dndphsym)+15*dndn./(8.*n.^2));
    end

    function[c6]=C6()
    c6=sqrt(n).*exp(-1i*ph-0.5*dphdph).*(-1i*dphdph+1./(2.*n).*(dndphsym+0.5i).*(1-dphdph)+1i./(8.*n.^2).*(2.*(dndphsym+0.5i).^2+dndn.*dphdph));        
    end

    function[c7]=C7()
c7=sqrt(n).*exp(-1i*ph-0.5*dphdph).*(1./(2.*n).*(1-2i*dndphsym)-dndn./(2.*n.^2));        
    end  
    
    
end