function [resultarray,timearray]=TruncatedWignerSolver(F,U,Delta,gamma,twinit,timearray)
%TRUNCATEDWIGNERSOLVER evolves trajectories with an XP-Gaussian
%ansatz in a vectorized way.
%F,U,Delta,gamma: physical parameters of the dynamics as defined in the main text
%twinit: vector of initial conditions for the field in Wigner phase-space
%(symmetrically ordered)
%timearray: row vector of the times when results must be stored
%resultarray: the phase-spase field values for different samples at
%different times.

% Initial conditions must be randomly sampled to describe a physical state,
% e.g. for aan initial coherent state with mean value alphainit, the input
% for a simulation of NTWsamples stochastic realisations can be obtained by
% Xrand=0.5*randn(1,NTWsamples)+real(alphainit); 
% Yrand=0.5*randn(1,NTWsamples)+imag(alphainit);
% twinit=Xrand+1i*Yrand;

%by W. Verstraelen, Theory of Quantum Systems & Complex Systems, dept.
%Physics, University of Antwerp for MATLAB 2017b.
%developed for 'Gaussian Quantum Trajectories for the Variational
%Simulation of Open Quatum-Optical Systems--WV and M. Wouters'



dt=1e-6;%1e-5%1e-4
%rng(20);
withGPU=false; %perform the calculation on the GPU?

if isrow(twinit), twinit=twinit.';end


resultarray=zeros(length(twinit),length(timearray));

if (withGPU)
    time=timearray(1);
    phi=gpuArray(twinit);
    vecsize=length(twinit);
    dens=phi.*conj(phi)-0.5;
    resultarray(:,1)=gather(phi);

    while time<timearray(end)
    ham=(-Delta-1i*gamma/2)+U.*(dens+1/2-1);
    phi=(1-1i*ham*dt).*phi+(-1i)*F*dt+sqrt(gamma*dt/4)*(gpuArray.randn(vecsize,1)+1i*gpuArray.randn(vecsize,1));
  
    dens=phi.*conj(phi)-0.5;

    time=time+dt;

    step=find(abs(timearray-time)<dt/2);
    if ~isempty(step)
        resultarray(:,step)=gather(phi);
    end

    end

else
    
    time=timearray(1);
    phi=twinit;
    vecsize=length(twinit);
    dens=phi.*conj(phi)-0.5;
    resultarray(:,1)=phi;

    while time<timearray(end)
    ham=(-Delta-1i*gamma/2)+U.*(dens+1/2-1);

    phi=(1-1i*ham*dt).*phi+(-1i)*F*dt+sqrt(gamma*dt/4)*(randn(vecsize,1)+1i*randn(vecsize,1));

    dens=phi.*conj(phi)-0.5;

    time=time+dt;

    %store the results if necessary
    step=find(abs(timearray-time)<dt/2);
    if ~isempty(step)
        resultarray(:,step)=phi;
    end

    end
    
    
    
    
end



end