function [MFresultarray,ADresultarray,Dresultarray,timearray]=XP_GaussianHeterodyneSolver(F,U,Delta,gamma,MFinit,ADinit,timearray)

%XP_GaussianHeterodyneSolver evolves trajectories with an XP-Gaussian
%ansatz in a vectorized way.
%F,U,Delta,gamma: physical parameters of the dynamics as defined in the main text
%MFinit: column vector of initial (complex) mean-field values 
%ADinit: column vector of initial anomalous densities of (complex) fluctuations (<\delta \delta>)
%timearray: row vector of the times when results must be stored
%MFresultarray: mean-field values for the different samples at different timesteps
%ADresultarray: anomalous densities of fluctuations <\delta \delta> for the different samples at different timesteps
%Dresultarray: densities of fluctuation <delta^\dagger delta> for the different samples at different timesteps

%by W. Verstraelen, Theory of Quantum Systems & Complex Systems, dept.
%Physics, University of Antwerp for MATLAB 2017b.
%developed for 'Gaussian Quantum Trajectories for the Variational
%Simulation of Open Quatum-Optical Systems--WV and M. Wouters'



dt=1e-6;%1e-5;%1e-4;
%rng(10); %random seed can be set if wanted
withGPU=false; %perform the calculation on the GPU?
assert(min(size(MFinit)==size(ADinit)));
assert(iscolumn(MFinit));%The different samples are ordered in the column dimension here.
vecsize=length(MFinit);

MFresultarray=zeros(vecsize,length(timearray)); %Mean-Field solutions
ADresultarray=zeros(vecsize,length(timearray)); %Anomalous density of the fluctuations <delta delta>
Dresultarray=zeros(vecsize,length(timearray)); %Density of the
%fluctuations <delta' delta> 


if(withGPU)

    time=timearray(1);
    alpha=gpuArray(MFinit);
    dd=gpuArray(ADinit);
    dens=-0.5+0.5*sqrt(1+4*abs(dd).^2);
    MFresultarray(:,1)=gather(alpha);
    ADresultarray(:,1)=gather(dd);
    Dresultarray(:,1)=gather(dens);

    sqdt=sqrt(dt);
    unit=gpuArray.ones(vecsize,1);
    sqgamotwo=sqrt(0.5*gamma);
    while time<timearray(end)

        conjalpha=conj(alpha);    
        dW1=sqdt*gpuArray.randn(vecsize,1);
        dW2=sqdt*gpuArray.randn(vecsize,1);
        d_alpha_det=(-0.5*gamma+1i*Delta)*alpha-1i*U*(conjalpha.*alpha.^2+2*alpha.*dens+conjalpha.*dd)-1i*F*unit;
        d_alpha_stochx=sqgamotwo*(dens+dd);
        d_alpha_stochy=1i*sqgamotwo*(dens-dd);
        d_dd=-1i*U*(alpha.^2.*(unit+2*dens)+dd.*(unit+4.*conjalpha.*alpha+6*dens))+(2i*Delta-gamma)*dd-2*gamma*dd.*dens;
        alpha=alpha+d_alpha_det*dt+d_alpha_stochx.*dW1+d_alpha_stochy.*dW2;
        dd=dd+d_dd*dt;
        dens=-0.5*unit+0.5*sqrt(unit+4*abs(dd).^2);
        time=time+dt;

        step=find(abs(timearray-time)<dt/2);
        if ~isempty(step)
            MFresultarray(:,step)=gather(alpha);
            ADresultarray(:,step)=gather(dd);
            Dresultarray(:,step)=gather(dens);
            assert(all(isfinite(alpha(:))));
        end

    end

else
    
    time=timearray(1);
    alpha=MFinit;
    dd=ADinit;
    dens=-0.5+0.5*sqrt(1+4*abs(dd).^2);
    MFresultarray(:,1)=alpha;
    ADresultarray(:,1)=dd;
    Dresultarray(:,1)=dens;

    sqdt=sqrt(dt);
    unit=ones(vecsize,1);
    sqgamotwo=sqrt(0.5*gamma);
    while time<timearray(end)

        conjalpha=conj(alpha);    
        dW1=sqdt*randn(vecsize,1);
        dW2=sqdt*randn(vecsize,1);
        d_alpha_det=(-0.5*gamma+1i*Delta)*alpha-1i*U*(conjalpha.*alpha.^2+2*alpha.*dens+conjalpha.*dd)-1i*F*unit; %eq. (16) 
        d_alpha_stochx=sqgamotwo*(dens+dd);
        d_alpha_stochy=1i*sqgamotwo*(dens-dd);
        d_dd=-1i*U*(alpha.^2.*(unit+2*dens)+dd.*(unit+4.*conjalpha.*alpha+6*dens))+(2i*Delta-gamma)*dd-2*gamma*dd.*dens; %eq. (11)
        alpha=alpha+d_alpha_det*dt+d_alpha_stochx.*dW1+d_alpha_stochy.*dW2;
        dd=dd+d_dd*dt;
        dens=-0.5*unit+0.5*sqrt(unit+4*abs(dd).^2); %from eq. (17)
        time=time+dt;

        % save result if necessary
        step=find(abs(timearray-time)<dt/2); 
        if ~isempty(step)
            MFresultarray(:,step)=alpha;
            ADresultarray(:,step)=dd;
            Dresultarray(:,step)=dens;
            assert(all(isfinite(alpha(:))));
            %[dW1 dW2]/sqdt ->uncomment for the random numbers to be
            %printed
        end

    end

end


end