function [MFresultarray,ADresultarray,Dresultarray,timearray]=XP_GaussianPhotonCountingSolver(F,U,Delta,gamma,MFinit,ADinit,timearray)

%XP_GaussianPhotonCountingSolver evolves trajectories with an XP-Gaussian
%ansatz in a vectorized way.
%F,U,Delta,gamma: physical parameters of the dynamics as defined in the main text
%MFinit: column vector of initial (complex) mean-field values 
%ADinit: column vector of initial anomalous densities of (complex) fluctuations (<\delta \delta>)
%timearray: row vector of the times when results must be stored
%MFresultarray: mean-field values for the different samples at different timesteps
%ADresultarray: anomalous densities of fluctuations <\delta \delta> for the different samples at different timesteps
%Dresultarray: densities of fluctuation <delta^\dagger delta> for the different samples at different timesteps

%by W. Verstraelen, Theory of Quantum Systems & Complex Systems, dept.
%Physics, University of Antwerp for MATLAB 2017b.
%developed for 'Gaussian Quantum Trajectories for the Variational
%Simulation of Open Quatum-Optical Systems--WV and M. Wouters'



dt=1e-6;%1e-5;%1e-4;
%rng(10); %random seed can be set if wanted
withGPU=false; %perform the calculation on the GPU?
assert(min(size(MFinit)==size(ADinit)));
assert(iscolumn(MFinit));%The different samples are ordered in the column dimension here.
vecsize=length(MFinit);

MFresultarray=zeros(vecsize,length(timearray)); %Mean-Field solutions
ADresultarray=zeros(vecsize,length(timearray)); %Anomalous density of the fluctuations <delta delta>
Dresultarray=zeros(vecsize,length(timearray)); %Density of the


if(withGPU) 

 time=timearray(1);
    alpha=gpuArray(MFinit);
    dd=gpuArray(ADinit);
    dens=-0.5+0.5*sqrt(1+4*abs(dd).^2);
    MFresultarray(:,1)=gather(alpha);
    ADresultarray(:,1)=gather(dd);
    Dresultarray(:,1)=gather(dens);

    
    unit=gpuArray.ones(vecsize,1);
    norm=unit;
 
    R=gpuArray.rand(vecsize,1);%Initialize the values of norm when there will be a jump    
   
 
    while time<timearray(end)

        conjalpha=conj(alpha);    
        totdens=alpha.*conjalpha+dens;

        
        %Is there a jump?
        jumppresent=(norm<R); nojumppresent=~jumppresent;
  
        
        R=nojumppresent.*R+jumppresent.*gpuArray.rand(vecsize,1);  
        
        
        
        alpha_jump=(conjalpha.*alpha.^2+2*alpha.*dens+conjalpha.*dd)./totdens; alpha_jump(isnan(alpha_jump))=0; %prevent tne NaN output for hypothetical jumps from zero density.
        dd_jump=((conjalpha.*alpha).^2.*dd+2*conjalpha.*alpha.*dens.*dd-alpha.^2.*dens.^2+3*dens.^2.*dd-conjalpha.^2.*dd.^2)./totdens.^2; dd_jump(isnan(dd_jump))=0;
        norm_jump=unit;
        
        alpha=nojumppresent.*alpha+jumppresent.*alpha_jump; conjalpha=conj(alpha);
        dd=nojumppresent.*dd+jumppresent.*dd_jump;
        norm=nojumppresent.*norm+jumppresent.*norm_jump;
        dens=-0.5+0.5*sqrt(1+4*abs(dd).^2); totdens=conjalpha.*alpha+dens;

        %Regardless of whether a jump happened, this is the deterministic
        %evolution.
        d_alpha=(-0.5*gamma+1i*Delta)*alpha-1i*U*(conjalpha.*alpha.^2+2*alpha.*dens+conjalpha.*dd)-1i*F*unit-gamma*(alpha.*dens+conjalpha.*dd);
        d_dd=-1i*U*(alpha.^2.*(unit+2*dens)+dd.*(unit+4.*conjalpha.*alpha+6*dens))+(2i*Delta-gamma)*dd-2*gamma*dd.*dens;
        d_norm=-gamma*totdens.*norm;
        
        alpha=alpha+d_alpha*dt;
        dd=dd+d_dd*dt;
        norm=norm+d_norm*dt;
        dens=-0.5+0.5*sqrt(1+4*abs(dd).^2); 

        time=time+dt;

        % save result if necessary
        step=find(abs(timearray-time)<dt/2);
        if ~isempty(step)
            MFresultarray(:,step)=gather(alpha);
            ADresultarray(:,step)=gather(dd);
            Dresultarray(:,step)=gather(dens);
            assert(all(isfinite(alpha(:))));
        end

    end
   
else
    
    time=timearray(1);
    alpha=MFinit;
    dd=ADinit;
    dens=-0.5+0.5*sqrt(1+4*abs(dd).^2);
    MFresultarray(:,1)=alpha;
    ADresultarray(:,1)=dd;
    Dresultarray(:,1)=dens;

    
    unit=ones(vecsize,1);%ones(vecsize,1);
    norm=unit;
 
    R=rand(vecsize,1);%Initialize the values of norm when there will be a jump    
   
 
    while time<timearray(end)

        conjalpha=conj(alpha);    
        totdens=alpha.*conjalpha+dens;

        
        %Is there a jump?
        jumppresent=(norm<R); nojumppresent=~jumppresent;        
        R=nojumppresent.*R+jumppresent.*rand(vecsize,1);
    
        
        alpha_jump=(conjalpha.*alpha.^2+2*alpha.*dens+conjalpha.*dd)./totdens; alpha_jump(isnan(alpha_jump))=0; %prevent tne NaN output for hypothetical jumps from zero density.
        dd_jump=((conjalpha.*alpha).^2.*dd+2*conjalpha.*alpha.*dens.*dd-alpha.^2.*dens.^2+3*dens.^2.*dd-conjalpha.^2.*dd.^2)./totdens.^2; dd_jump(isnan(dd_jump))=0;
        norm_jump=unit;
        
        alpha=nojumppresent.*alpha+jumppresent.*alpha_jump; conjalpha=conj(alpha);
        dd=nojumppresent.*dd+jumppresent.*dd_jump;
        norm=nojumppresent.*norm+jumppresent.*norm_jump;
        dens=-0.5+0.5*sqrt(1+4*abs(dd).^2); totdens=conjalpha.*alpha+dens;

        %Regardless of whether a jump happened, this is the deterministic
        %evolution.
        d_alpha=(-0.5*gamma+1i*Delta)*alpha-1i*U*(conjalpha.*alpha.^2+2*alpha.*dens+conjalpha.*dd)-1i*F*unit-gamma*(alpha.*dens+conjalpha.*dd);
        d_dd=-1i*U*(alpha.^2.*(unit+2*dens)+dd.*(unit+4.*conjalpha.*alpha+6*dens))+(2i*Delta-gamma)*dd-2*gamma*dd.*dens;
        d_norm=-gamma*totdens.*norm;
        
        alpha=alpha+d_alpha*dt;
        dd=dd+d_dd*dt;
        norm=norm+d_norm*dt;
        dens=-0.5+0.5*sqrt(1+4*abs(dd).^2); 

        time=time+dt;

        % save result if necessary
        step=find(abs(timearray-time)<dt/2);
        if ~isempty(step)
            MFresultarray(:,step)=alpha;
            ADresultarray(:,step)=dd;
            Dresultarray(:,step)=dens;
            assert(all(isfinite(alpha(:))));
        end

    end

    
end

end